/* Distributed Checksum Clearinghouse
 *
 * control dcc server
 *
 * Copyright (c) 2011 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software without
 * changes for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear in all
 * copies and any distributed versions or copies are either unchanged
 * or not called anything similar to "DCC" or "Distributed Checksum
 * Clearinghouse".
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC by contacting Rhyolite Software
 * at sales@rhyolite.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.141-1.264 $Revision$
 */

#include "dcc_ck.h"
#include "dcc_xhdr.h"
#include "dcc_heap_debug.h"
#include "dcc_ids.h"
#ifndef DCC_WIN32
#include <arpa/inet.h>
#endif

static DCC_EMSG dcc_emsg;
static DCC_FNM_LNO_BUF fnm_buf;

static DCC_CLNT_CTXT *ctxt;

static DCC_PATH info_map_nm = DCC_MAP_NM_DEF;
static u_char dup_emsg;
static const char *ids_nm;
static time_t clock_kludge;
static const char *homedir;
static DCC_PASSWD passwd;
static u_char passwd_set;
static DCC_IP src;
static DCC_SRVR_NM srvr = DCC_SRVR_NM_DEF(DCC_SRVR_PORT_HO);
static u_char port_set;
static DCC_CLNT_ID srvr_clnt_id = DCC_ID_ANON;
static enum WHICH_MAP {MAP_TMP, MAP_INFO} which_map = MAP_INFO;
static u_char map_changed = 1;

static DCC_INFO_FGS info_fgs;

static u_char grey_set;
static DCC_CLNT_FGS grey_clnt_fgs;

static u_char quiet;


static u_char do_cmds(char *);
static void set_which_map(enum WHICH_MAP);
static u_char init_map(u_char, u_char);

struct cmd_tbl_entry;
/* -1=display help message, 0=command failed, 1=success */
typedef int CMD (const char *, const struct cmd_tbl_entry *);
typedef struct cmd_tbl_entry {
    const char	*cmd;
    CMD		(*fnc);
    u_char	args;			/* 0=optional, 1=required, 2=none */
    u_char	privileged;		/* 1=must have server's password */
    u_char	write_map;		/* 1=write map, 2=write /var/dcc/map */
    const char	*help_str;
} CMD_TBL_ENTRY;

static CMD help_cmd;
static CMD exit_cmd;
static CMD grey_cmd;
static CMD file_cmd;
static CMD new_map_cmd;
static CMD delete_cmd;
static CMD add_cmd;
static CMD load_cmd;
static CMD host_cmd;
static CMD port_cmd;
static CMD passwd_cmd;
static CMD id_cmd;
static CMD homedir_cmd;
#ifndef DCC_WIN32
static CMD libexecdir_cmd;
#endif
static CMD debug_cmd;
static CMD quiet_cmd;
static CMD no_fail_cmd;
static CMD ckludge_cmd;
static CMD ipv6_cmd;
static CMD src_cmd;
static CMD socks_cmd;
static CMD info_cmd;
static CMD rtt_cmd;
static CMD delck_cmd;
static CMD sleep_cmd;
static CMD clients_cmd;
static CMD anon_cmd;
static CMD flod_rewind;
static CMD ffwd_in;
static CMD ffwd_out;
static CMD flod_stats;
static CMD flod_stats_clear;
static CMD stats_cmd;
static CMD clock_ck_cmd;
static CMD trace_def;

static const CMD_TBL_ENTRY cmds_tbl[] = {
    {"help",		help_cmd,   0, 0, 0, "help [cmd]"},
    {"?",		0,	    0, 0, 0, 0},
    {"exit",		exit_cmd,   2, 0, 0, "exit"},
    {"quit",		0,	    2, 0, 0, 0},
    {"grey",		grey_cmd,   0, 0, 0, "grey [on|off]"},
    {"homedir",		homedir_cmd,0, 0, 0, "homedir [path]"},
#ifndef DCC_WIN32
    {"libexecdir",   libexecdir_cmd,2, 0, 0, "libexecdir"},
#endif
    {"file",		file_cmd,   0, 0, 0, "file [mapfile]"},
    {"map",		0,	    0, 0, 0, 0},
    {"new map",		new_map_cmd,0, 0, 0, "new map [mapfile]"},
    {"delete",		delete_cmd, 1, 0, 2, "delete host[,port]"},
    {"add",		add_cmd,    1, 0, 2,
	    "add host,[port|-] [RTT+/-#] [ID [passwd]]"},
    {"load",		load_cmd,   1, 0, 2, "load {map.txtfile | -}"},
    {"host",		host_cmd,   0, 0, 0, "host [name]"},
    {"server",		0,	    0, 0, 0, 0},
    {"port",		port_cmd,   0, 0, 0, "port #"},
    {"password",	passwd_cmd, 0, 0, 0, "password secret"},
    {"passwd",		0,	    0, 0, 0, 0},
    {"id",		id_cmd,	    0, 0, 0, "id [ID]"},
    {"debug",		debug_cmd,  0, 0, 0, "debug [on|off|TTL=x|#]"},
    {"quiet",		quiet_cmd,  0, 0, 0, "quiet [on|off]"},
    {"no fail",		no_fail_cmd,2, 0, 2, "no fail"},
    {"clock kludge",	ckludge_cmd,0, 0, 0, "clock kludge +/-secs"},
    {"IPv6",		ipv6_cmd,   0, 0, 0, "IPv6 [on|off|only]"},
    {"src",		src_cmd,    0, 0, 0, "src [-|IPaddress]"},
    {"SOCKS",		socks_cmd,  0, 0, 0, "SOCKS [on|off]"},
    {"info",		info_cmd,   0, 0, 0, "info [-N]"},
    {"RTT",		rtt_cmd,    0, 0, 0, "RTT [-N]"},
    {"delck",		delck_cmd,  1, 1, 0, "delck type hex1..4"},
    {"sleep",		sleep_cmd,  1, 0, 0, "sleep sec.onds"},
    {"clients",		clients_cmd,0, 0, 0,
	    "clients [-nsiaVAK] [max [thold [addr/prefix]]]"},
    {"anon delay",	anon_cmd,   0, 0, 0, "\nanon delay [delay[,inflate]]"},
    {"flood rewind",	flod_rewind,1, 1, 0, "flood rewind ID"},
    {"flod rewind",	0,	    1, 1, 0, 0},
    {"flood FFWD in",	ffwd_in,    1, 1, 0, "flood FFWD in ID"},
    {"flod FFWD in",	0,	    1, 1, 0, 0},
    {"flood FFWD out",	ffwd_out,   1, 1, 0, "flood FFWD out ID"},
    {"flod FFWD out",	0,	    1, 1, 0, 0},
    {"flood stats clear", flod_stats_clear,
				    1, 1, 0, "flood stats [clear] {ID|all}"},
    {"flood status clear", 0,	    1, 1, 0, 0},
    {"flod stats clear",  0,	    1, 1, 0, 0},
    {"flod status clear", 0,	    1, 1, 0, 0},
    {"flood stats",	flod_stats, 1, 0, 0, 0},
    {"flood status",	0,	    1, 0, 0, 0},
    {"flod stats",	0,	    1, 0, 0, 0},
    {"flod status",	0,	    1, 0, 0, 0},
    {"stats",		stats_cmd,  0, 0, 0, "stats [clear|all]"},
    {"status",		0,	    0, 0, 0, 0},
    {"clock check",	clock_ck_cmd,0,0, 0, "clock check"},
    {"trace default",	trace_def,  2, 1, 0, "trace default"},
};


#define PRV_MSG ";\n"		\
"   use the \"id server-ID\" command\n"	\
"   and either \"password secret\" command or `su` to read passwords from %s"


static DCC_OP_RESP aop_resp;
static struct timeval op_start, op_end;
static DCC_SOCKU op_result_su;

static struct {
    const char	*op;
    const char	*help_str;
    DCC_AOPS	aop;
    u_char	privileged;
    u_int32_t	val;
} aops_tbl[] = {
#define TMAC(s,b) \
    {"trace "#s" on",	"trace "#s" {on|off}",			\
			    DCC_AOP_TRACE_ON, 1, DCC_TRACE_##b},\
    {"trace "#s" off",	0, DCC_AOP_TRACE_OFF,	1, DCC_TRACE_##b}
    TMAC(admn,ADMN_BIT),
    TMAC(anon,ANON_BIT),
    TMAC(clnt,CLNT_BIT),
    TMAC(rlim,RLIM_BIT),
    TMAC(query,QUERY_BIT),
    TMAC(ridc,RIDC_BIT),
    TMAC(flood,FLOD_BIT),
    TMAC(flood2,FLOD2_BIT),
    /* TMAC(xx,_BIT), */
    TMAC(bl,BL_BIT),
    TMAC(db,DB_BIT),
    TMAC(wlist,WLIST_BIT),
#undef TMAC

    {"stop",		"",	DCC_AOP_STOP,	    1, 0},
    {"system stop",	"",	DCC_AOP_STOP,	    1, 1},
    {"clean stop",	"",	DCC_AOP_STOP,	    1, 2},
    {"flood check",	"",	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_CHECK},
    {"flod check",	0,	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_CHECK},
    {"flood shutdown",	"",	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_SHUTDOWN},
    {"flod shutdown",	0,	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_SHUTDOWN},
    {"flood halt",	"",	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_HALT},
    {"flod halt",	0,	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_HALT},
    {"flood resume",	"",	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_RESUME},
    {"flod resume",	0,	DCC_AOP_FLOD,	    1, DCC_AOP_FLOD_RESUME},
    {"flood list",	"",	DCC_AOP_FLOD,	    0, DCC_AOP_FLOD_LIST},
    {"flod list",	0,	DCC_AOP_FLOD,	    0, DCC_AOP_FLOD_LIST},
    {"DB clean",	"",	DCC_AOP_DB_CLEAN,   1, 0},
    {"DB new",		"",	DCC_AOP_DB_NEW,	    1, 0},
    {"DB flush cache",	"",	DCC_AOP_DB_UNLOAD,  1, 0},
    {"DB cache ok",	"",	DCC_AOP_DB_UNLOAD,  1, 1},
};


static void DCC_NORET
usage(void)
{
	dcc_logbad(EX_USAGE,
		   "usage: [-VBdq] [-h homedir] [-c ids] [op1 [op2] ... ]\n");
}



static void DCC_PF(1,2)
cdcc_error_msg(const char *p, ...)
{
	va_list args;

	if (dup_emsg) {
		va_start(args, p);
		vprintf(p, args);
		va_end(args);
	}

	/* on some systems cannot use args twice after 1 va_start() */
	va_start(args, p);
	dcc_verror_msg(p, args);
	va_end(args);
}



int
main(int argc, char **argv)
{
	char cmd_buf[500];
	int i;

	dcc_init_priv();
	dcc_syslog_init(0, argv[0], 0);

	while ((i = getopt(argc, argv, "VBdqh:c:")) != -1) {
		switch (i) {
		case 'V':
			fprintf(stderr, DCC_VERSION"\n");
			break;

		case 'B':
			dup_emsg = 1;
			break;

		case 'd':
			++dcc_clnt_debug;
			break;

		case 'q':
			++quiet;
			break;

		case 'h':
			homedir = optarg;
			break;

		case 'c':
			ids_nm = optarg;
			break;

		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	dcc_clnt_unthread_init();
	if (!dcc_cdhome(0, homedir, 1) && homedir)
		dcc_cdhome(0, homedir = 0, 1);
	set_ids_path(0, ids_nm);
	wf_init(&cmn_wf, 0);

	dcc_all_srvrs = 1;

	if (!init_map(!quiet, 0))
		set_which_map(MAP_TMP);

	/* with a list of commands, act as a batch utility */
	if (argc != 0) {
		for (;;) {
			/* a final arg of "-" says switch to interactive mode */
			if (argc == 1 && !strcmp(*argv, "-"))
				break;

			if (!do_cmds(*argv)) {
				fputs(" ?\n", stderr);
				exit(EX_UNAVAILABLE);
			}
			assert_ctxts_unlocked();
			assert_info_unlocked();

			++argv;
			if (!--argc) {
				exit(EX_OK);
			}
		}
	}

	/* Without an arg list of commands, look for commands from STDIN.
	 * Commands end with a semicolon or newline. */
	for (;;) {
		assert_ctxts_unlocked();
		assert_info_unlocked();
		printf("cdcc %s> ",
		       which_map == MAP_INFO ? info_map_nm : "-");
		fflush(stderr);
		fflush(stdout);
		if (!fgets(cmd_buf, sizeof(cmd_buf), stdin)) {
			fputc('\n', stdout);
			fflush(stderr);
			exit(EX_OK);
		}
		if (!do_cmds(cmd_buf)) {
			fputs(" ?\n", stderr);
			fflush(stderr);
		}
	}
}



/* see if don't need a server-ID password, have one or if we can get it */
static u_char				/* 0=failed, 1=ok */
get_passwd(u_char privileged)		/* 0=no privileges need, 1=need power */
{
	const ID_TBL *srvr_clnt_tbl;
	DCC_PATH tmp;

	srvr.clnt_id = srvr_clnt_id;
	if (passwd_set) {
		/* set to use the manual password */
		memcpy(srvr.passwd, passwd, sizeof(srvr.passwd));
		/* succeed if this is not a priviledge command
		 * or if we won't be trying to get the server to do
		 * something priviledge for the anonymous client-ID */
		return (!privileged || srvr_clnt_id != DCC_ID_ANON);
	}

	memset(srvr.passwd, 0, sizeof(srvr.passwd));

	/* fail if we would be trying to get the server to do something
	 * powerful for the anonymous client-ID */
	if (srvr_clnt_id == DCC_ID_ANON)
		return !privileged;

	/* Fetch the common server passwords only if we can read them without
	 * set-UID.  This keeps random local users from attacking local
	 * or remote servers with privileged commands, but does not slow
	 * down privilege users who could use an editor to read and use
	 * the cleartext passwords manually. */
	dcc_rel_priv();
	if (0 > access(ids_path, R_OK) && errno == EACCES) {
		srvr.clnt_id = DCC_ID_ANON;
		if (privileged)
			dcc_error_msg("access(%s): %s",
				      dcc_fnm2abs_msg(tmp, ids_path),
				      ERROR_STR());
		return !privileged;
	}

	if (0 >= load_ids(dcc_emsg, srvr_clnt_id, &srvr_clnt_tbl, 1)) {
		/* complain about problems in the file if we need it */
		if (privileged)
			dcc_error_msg("%s", dcc_emsg);
		if (!srvr_clnt_tbl) {
			srvr.clnt_id = DCC_ID_ANON;
			return !privileged;
		}
	}

	memcpy(srvr.passwd, srvr_clnt_tbl->cur_passwd,
	       sizeof(srvr.passwd));
	return 1;
}



static void
set_which_map(enum WHICH_MAP new)
{
	/* release things even if nothing seems to be changing
	 * to ensure that we bind a new socket and open a new map file */
	dcc_ctxts_lock();
	if (dcc_clnt_info)
		dcc_unmap_close_info(0);
	if (ctxt) {
		dcc_rel_ctxt(ctxt);
		ctxt = 0;
	}
	dcc_ctxts_unlock();

	map_changed = 1;
	which_map = new;
	if (new == MAP_INFO) {
		passwd_set = 0;
		src.family = AF_UNSPEC;
	}
}



static u_char
cdcc_unlock(u_char complain)
{
	u_char result;

	result = dcc_info_unlock(dcc_emsg);
	dcc_ctxts_unlock();
	if (!result && complain)
		dcc_error_msg("%s", dcc_emsg);
	return result;
}



/* start talking to the local map file */
static u_char				/* 0=failed 1=mapped and locked */
init_map(u_char complain,
	 u_char lock)			/* 1=keep both locks on success */
{
	u_char result;

	dcc_emsg[0] = '\0';
	dcc_ctxts_lock();
	if (which_map == MAP_TMP) {
		result = (dcc_map_tmp_info(dcc_emsg, &srvr, &src, info_fgs)
			  && dcc_info_lock(dcc_emsg));
	} else {
		result = dcc_map_lock_info(dcc_emsg, info_map_nm, -1);
	}
	if (result) {
		info_fgs = dcc_clnt_info->fgs;
		if (!lock)
			result = cdcc_unlock(complain);
	} else {
		dcc_ctxts_unlock();
		if (complain)
			dcc_error_msg("%s", dcc_emsg);
	}
	return result;
}



/* get ready to start talking to a DCC server
 *	on entry only the contexts are locked and only if we have a
 *	    context and the file is not mapped
 *	both locks are released on exit */
static u_char				/* 1=ok, 0=failed */
rdy_ctxt(DCC_CLNT_FGS fgs)
{
	u_char rdy_done;		/* 1=readied things enough */
	u_char need_unlock;
	u_char result;

	rdy_done = 0;
	need_unlock = 0;
	result = 0;

	if (grey_on)
		fgs |= DCC_CLNT_FG_GREY;
	else
		fgs &= ~DCC_CLNT_FG_GREY;
	fgs |= DCC_CLNT_FG_NO_FAIL;

	if (!dcc_clnt_info && ctxt) {
		dcc_rel_ctxt(ctxt);
		ctxt = 0;
	}

	if (!ctxt) {
		if (which_map == MAP_TMP) {
			/* create a new temporary map */
			ctxt = dcc_tmp_clnt_init(dcc_emsg, ctxt, &srvr, &src,
						 fgs, info_fgs);
		} else {
			/* open official map file */
			ctxt = dcc_clnt_init(dcc_emsg, ctxt, info_map_nm, fgs);
		}
		if (!ctxt) {
			dcc_error_msg("%s", dcc_emsg);
			return 0;
		}
		rdy_done = 1;
		result = 1;
	}

	/* if there is only a greylist server name, care only about it */
	if (!grey_set && dcc_clnt_info
	    && !grey_on
	    && dcc_clnt_info->dcc.nms[0].hostname[0] == '\0'
	    && dcc_clnt_info->grey.nms[0].hostname[0] != '\0') {
		grey_on = 1;
		fgs |= DCC_CLNT_FG_GREY;
		rdy_done = 0;
		result = 0;
	}

	dcc_ctxts_lock();
	if (!rdy_done) {
		dcc_emsg[0] = '\0';
		if (dcc_clnt_rdy(dcc_emsg, ctxt, fgs)) {
			need_unlock = 1;
			result = 1;
		} else {
			dcc_error_msg("%s", dcc_emsg);
			result = 0;
		}
		/* we must give up now if we do not have the file */
		if (!dcc_clnt_info) {
			dcc_rel_ctxt(ctxt);
			ctxt = 0;
			dcc_ctxts_unlock();
			dcc_error_msg("%s", dcc_emsg);
			return 0;
		}
	}
	info_fgs = dcc_clnt_info->fgs;
	if (!(fgs & DCC_CLNT_FG_NO_MEASURE_RTTS))
		map_changed = 0;

	if (need_unlock && !dcc_info_unlock(0)) {
		dcc_rel_ctxt(ctxt);
		ctxt = 0;
		dcc_ctxts_unlock();
		dcc_error_msg("%s", dcc_emsg);
		return 0;
	}

	/* check the other (greylist or not) server */
	if (which_map == MAP_INFO) {
		dcc_emsg[0] = '\0';
		if (!dcc_clnt_rdy(dcc_emsg, ctxt, fgs ^ DCC_CLNT_FG_GREY)) {
			if (dcc_clnt_debug > 1)
				dcc_error_msg("%s", dcc_emsg);
		} else {
			dcc_info_unlock(0);
		}
	}

	dcc_ctxts_unlock();
	return result;
}



/* should hold both locks on entry; both are released on exit */
static void
fix_info(void)
{
	map_changed = 1;

#ifdef DCC_DEBUG_CLNT_LOCK
	assert_ctxts_locked();
	assert_info_locked();
#endif

	dcc_force_measure_rtt(&dcc_clnt_info->dcc);
	dcc_force_measure_rtt(&dcc_clnt_info->grey);

	cdcc_unlock(1);

	/* repair addresses in the real map file now */
	if (!quiet
	    && which_map == MAP_INFO)
		rdy_ctxt(DCC_CLNT_FG_BAD_SRVR_OK);
}



/* compare ignoring case */
static const char *			/* 0 or mismatch in str */
cmd_cmp(const char *str,		/* from the user */
	const char *cmd,		/* command table entry */
	u_char *partial)		/* initial substring match */
{
	char op_c, str_c;
	int len;

	if (partial)
		*partial = 0;

	len = 0;
	for (;;) {
		op_c = *cmd;
		op_c = DCC_TO_LOWER(op_c);

		str_c = *str;
		if (str_c == '\t')
			str_c = ' ';
		else
			str_c = DCC_TO_LOWER(str_c);

		if (op_c != str_c) {
			/* compress bursts of blanks */
			if (str_c == ' ' && len != 0 && *(cmd-1) == ' ') {
				++str;
				continue;
			}

			/* Treat substring match without an arg
			 * as a complete match. */
			if (str_c == '\0') {
				if (partial)
					*partial = 1;
				return 0;
			}
			return str;
		}

		/* stop at an exact match */
		if (op_c == '\0')
			return 0;

		++cmd;
		++str;
		++len;
	}
}



/* Display our name for the server and its address,
 * while suppressing some duplicates */
static void
print_aop(SRVR_INX srvr_inx)		/* NO_SRVR or server index */
{
	const DCC_SRVR_CLASS *class;
	char date_buf[40];
	char sustr[DCC_SU2STR_SIZE];
	DCC_PATH tmp;
	const char *srvr_nm;
	NAM_INX nam_inx;

	dcc_su2str2(sustr, sizeof(sustr), &op_result_su);
	class = DCC_GREY2CLASS(grey_on);

	/* Display the currently chosen server
	 * if the specified srvr_inx is NO_SRVR */
	if (!VALID_SRVR(class, srvr_inx))
		srvr_inx = class->srvr_inx;

	if (VALID_SRVR(class, srvr_inx)
	    && (VALID_NAM(nam_inx = class->addrs[srvr_inx].nam_inx))) {
		srvr_nm = class->nms[nam_inx].hostname;
		if (strcmp(srvr_nm, sustr)
		    && strcmp(srvr_nm, DCC_LOCALHOST)) {
			fputs(srvr_nm, stdout);
			putchar(' ');
		}
		printf("%s\n        server-ID %d",
		       dcc_su2str_err(&op_result_su),
		       class->addrs[srvr_inx].srvr_id);
	} else {
		printf("%s\n                    ",
		       dcc_su2str_err(&op_result_su));
	}
	if (srvr.clnt_id != DCC_ID_ANON)
		printf("  client-ID %d", srvr.clnt_id);
	if (which_map == MAP_INFO)
		printf("  %s", dcc_fnm2abs_msg(tmp, info_map_nm));
	dcc_time2str(date_buf, sizeof(date_buf), "  %X",
		     op_start.tv_sec);
	fputs(date_buf, stdout);
	putchar('\n');
}



static u_char				/* 0=some kind of problem, 1=done */
start_aop(DCC_AOPS aop, u_int32_t val1, SRVR_INX srvr_inx)
{
	DCC_OPS result;

	if (!rdy_ctxt(0))
		return 0;

	gettimeofday(&op_start, 0);
	result = dcc_aop(dcc_emsg, ctxt, grey_clnt_fgs, srvr_inx, clock_kludge,
			 aop, val1, 0, 0, 0, 0, 0, &aop_resp, &op_result_su);
	gettimeofday(&op_end, 0);

	if (result == DCC_OP_INVALID
	    || result == DCC_OP_ERROR) {
		cdcc_error_msg("%s", dcc_emsg);
		return 0;
	}

	return 1;
}



static void
fin_aop(SRVR_INX srvr_inx,		/* index of server */
	u_char psrvr)			/* 1=print server name */
{
	if (quiet && !dcc_clnt_debug)
		return;

	if (psrvr)
		print_aop(srvr_inx);

	/* say what the server had to say */
	if (aop_resp.resp.val.string[0] >= ' '
	    && aop_resp.resp.val.string[0] < 0x7f) {
		fputs(aop_resp.resp.val.string, stdout);
		putchar('\n');
	}

	if (dcc_clnt_debug) {
		printf("%.2f ms\n",
		       ((op_end.tv_sec-op_start.tv_sec)*1000.0
			+ (op_end.tv_usec-op_start.tv_usec)/1000.0));
	}
	putchar('\n');
}



static u_char				/* 0=some kind of problem, 1=done */
do_aop(DCC_AOPS aop, u_int32_t val1, SRVR_INX srvr_inx, u_char psrvr)
{
	if (!start_aop(aop, val1, srvr_inx))
		return 0;
	fin_aop(srvr_inx, psrvr);
	return 1;
}



static u_char				/* 0=not enough power */
ck_cmd_priv(const CMD_TBL_ENTRY *ce,
	    u_char privileged,		/* 1=need good server-ID & password */
	    u_char write_map)		/* 1=write map, 2=write /var/dcc/map */
{
	DCC_PATH tmp;

	/* always call get_passwd() so we have always fetched a password
	 * fail if this command needs a good server-ID and password */
	if (!get_passwd(privileged)) {
		dcc_error_msg("\"%s\" is a privileged server command"PRV_MSG,
			      ce->cmd, ids_path);
		return 0;
	}

	if (!write_map)
		return 1;

	/* we can always write to our own throw-away map file */
	if (write_map == 1 && which_map == MAP_TMP)
		return 1;

	if (0 > access(info_map_nm, R_OK)
	    && errno != ENOENT && errno != ENOTDIR) {
		dcc_error_msg("\"%s\" is a privileged command changing %s",
			      ce->cmd, dcc_fnm2abs_msg(tmp, info_map_nm));
		return 0;
	}
	return 1;
}



static u_char				/* 1=ok 0=bad command */
cmd(const char *op)
{
	const char *arg, *help_str;
	int op_num, j;
	const CMD_TBL_ENTRY *ce;
	u_char partial;

	/* look for the string as a command and execute it if we find it */
	ce = &cmds_tbl[0];
	for (op_num = 0; op_num < DIM(cmds_tbl); ++op_num) {
		if (cmds_tbl[op_num].fnc)
			ce = &cmds_tbl[op_num];
		arg = cmd_cmp(op, cmds_tbl[op_num].cmd, &partial);
		/* if the command table entry and the command completely
		 * matched, then infer a null argument */
		if (!arg) {
			if (partial) {
				help_cmd(op, 0);
				return 0;
			}
			if (!ck_cmd_priv(ce, ce->privileged, ce->write_map))
				return 0;
			if (ce->args != 1) {
				j = ce->fnc("", ce);
				if (j >= 0)
					return j;
			}
			help_cmd(op, 0);
			return 0;
		}
		/* If the command table entry is an initial sustring of
		 * the user's command, then the rest of the command must
		 * start with white space or '='.  Allow '=' to let
		 * homedir/fix-map not need use `eval` to quote blanks
		 * `eval` in bash loses exit status.
		 *
		 * Trim blanks and use the rest of the string as the argument */
		j = strspn(arg, DCC_WHITESPACE"=");
		if (j) {
			if (ce->args == 2) {
				help_cmd(op, 0);    /* arg not allowed */
				return 0;
			}
			if (!ck_cmd_priv(ce, ce->privileged, ce->write_map))
				return 0;
			j = ce->fnc(arg+j, ce);
			if (j >= 0)
				return j;
			help_cmd(op, 0);
			return 0;
		}
	}


	/* otherwise try to interpret it as a DCC administrative packet */
	op_num = 0;
	help_str = "";
	for (;;) {
		if (op_num >= DIM(aops_tbl)) {
			dcc_error_msg("unrecognized command \"%s\"", op);
			return 0;
		}
		/* do a command */
		if (aops_tbl[op_num].help_str) {
			help_str = aops_tbl[op_num].help_str;
			if (*help_str == '\0')
				help_str = aops_tbl[op_num].op;
		}
		if (!cmd_cmp(op, aops_tbl[op_num].op, &partial)) {
			if (partial) {
				help_cmd(op, 0);
				return 0;
			}
			break;
		}
		++op_num;
	}

	/* send an administrative request to the server */
	if (!get_passwd(aops_tbl[op_num].privileged)) {
		dcc_error_msg("\"%s\" is a privileged operation"PRV_MSG,
			      help_str, ids_path);
		return 0;
	}

	/* try to send it */
	return do_aop(aops_tbl[op_num].aop, aops_tbl[op_num].val,
		      NO_SRVR, 1);
}



static u_char				/* 0=bad command, 1=ok */
do_cmds(char *cmd_buf)
{
	char *next_cmd, *cur_cmd, *cmd_end;
	char c;

	next_cmd = cmd_buf;
	for (;;) {
		cur_cmd = next_cmd + strspn(next_cmd, DCC_WHITESPACE";");

		if (*cur_cmd == '#' || *cur_cmd == '\0')
			return 1;

		next_cmd = cur_cmd + strcspn(cur_cmd, ";\n\r");
		cmd_end = next_cmd;
		next_cmd += strspn(next_cmd, ";\n\r");

		/* null terminate and trim trailing white space from
		 * command or arg */
		do {
			*cmd_end-- = '\0';
			c = *cmd_end;
		} while (cmd_end >= cur_cmd
			 && strchr(DCC_WHITESPACE";", c));

		if (*cur_cmd == '\0')	/* ignore blank commands */
			continue;

		if (!cmd(cur_cmd))
			return 0;
	}
}



static int
help_cmd_print(int pos, const char *str)
{
#define HELP_COL 24
	int col, nl;

	if (str[0] == '\n') {
		nl = 100;
		++str;
	} else {
		nl = 0;
	}
	col = strlen(str)+1;
	col += HELP_COL - (col % HELP_COL);
	pos += col;
	if (pos > 78) {
		putchar('\n');
		pos = col;
	}
	printf("%-*s", col, str);
	pos += nl;

	return pos;
#undef HELP_COL
}



static int
help_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	u_char found;
	const char *help_str;
	const char *p;
	u_char iss;
	int pos;
	int i;

	/* Say something about one command.
	 * If the string is a initial substring of at least one command,
	 * talk about only commands that also have the strings as an
	 * initial substrings */
	if (arg && *arg != '\0') {
		iss = 0;
		for (i = 0; i < DIM(cmds_tbl) && !iss; ++i) {
			if (!cmd_cmp(arg, cmds_tbl[i].cmd, 0))
				iss = 1;
		}
		for (i = 0; i < DIM(aops_tbl); ++i) {
			if (!cmd_cmp(arg, aops_tbl[i].op, 0))
				iss = 1;
		}
		found = 0;
		help_str = 0;
		for (i = 0; i < DIM(cmds_tbl); ++i) {
			if (cmds_tbl[i].help_str)
				help_str = cmds_tbl[i].help_str;
			if (!help_str)
				continue;
			p = cmd_cmp(arg, cmds_tbl[i].cmd, 0);
			if (!p || (!iss && (*p == ' ' || *p == '\t'))) {
				/* skip a leading '\n' that signals a
				 * new line after this command */
				if (*help_str == '\n')
					++help_str;
				printf("usage: %s\n", help_str);
				help_str = 0;
				found = 1;
			}
		}
		for (i = 0; i < DIM(aops_tbl); ++i) {
			/* note the first help string in a group of synonyms */
			if (aops_tbl[i].help_str) {
				help_str = aops_tbl[i].help_str;
				/* use the command itself if the string empty */
				if (*help_str == '\0')
					help_str = aops_tbl[i].op;
			}
			if (!help_str)
				continue;
			p = cmd_cmp(arg, aops_tbl[i].op, 0);
			if (!p || (!iss && (*p == ' ' || *p == '\t'))) {
				/* skip a leading '\n' that signals a
				 * new line after this command */
				if (*help_str == '\n')
					++help_str;
				printf("usage: %s\n", help_str);
				help_str = 0;
				found = 1;
			}
		}

		if (found)
			return 1;
	}

	/* talk about all of the commands */
	printf("   version "DCC_VERSION"\n");
	pos = 0;
	for (i = 0; i < DIM(cmds_tbl); ++i) {
		if (cmds_tbl[i].help_str)
			pos = help_cmd_print(pos, cmds_tbl[i].help_str);
	}
	for (i = 0; i < DIM(aops_tbl); ++i) {
		help_str = aops_tbl[i].help_str;
		if (!help_str)
			continue;
		if (*help_str == '\0')
			help_str = aops_tbl[i].op;
			pos = help_cmd_print(pos, help_str);
	}
	putchar('\n');

	return 1;
}



static int DCC_NORET
exit_cmd(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	exit(EX_OK);
#ifndef HAVE_GCC_ATTRIBUTES
	return -1;
#endif
}


static int
grey_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (arg[0] == '\0') {
		printf("    Greylist mode %s%s\n",
		       grey_on ? "on" : "off",
		       grey_set ? "" : " by default");
		return 1;
	}
	if (!strcmp(arg, "off")) {
		grey_on = 0;
		grey_set = 1;
		grey_clnt_fgs &= ~DCC_CLNT_FG_GREY;
		set_which_map(which_map);
	} else if (!strcmp(arg, "on")) {
		grey_on = 1;
		grey_set = 1;
		grey_clnt_fgs |= DCC_CLNT_FG_GREY;
		set_which_map(which_map);
	} else {
		return -1;
	}
	if (!port_set)
		srvr.port = DCC_GREY2PORT(grey_on);
	return 1;
}



static int
homedir_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (arg[0] != '\0') {
		if (!dcc_cdhome(0, arg, 1))
			return 0;
		if (ids_nm && !set_ids_path(dcc_emsg, ids_nm))
			dcc_error_msg("%s", dcc_emsg);
		set_which_map(MAP_INFO);
	}
	printf("    homedir=%s\n", dcc_homedir);
	return 1;
}



#ifndef DCC_WIN32
static int
libexecdir_cmd(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	printf("    libexecdir=%s\n", DCC_LIBEXECDIR);
	return 1;
}
#endif



/* set name of map file */
static int
file_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_PATH tmp;

	if (arg[0] == '\0') {
		if (which_map == MAP_INFO)
			printf("    using map file %s\n",
			       dcc_fnm2abs_msg(tmp, info_map_nm));
		else
			printf("    map file %s but using temporary file\n",
			       dcc_fnm2abs_msg(tmp, info_map_nm));
		return 1;
	}

	BUFCPY(info_map_nm, arg);
	set_which_map(MAP_INFO);
	return 1;
}



/* create a new client map or parameter file */
static int
new_map_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_PATH tmp;

	if (arg[0] == '\0')
		arg = DCC_MAP_NM_DEF;

	dcc_rel_priv();
	if (!dcc_create_map(dcc_emsg, arg, 0, 0, 0, 0, 0, 0, info_fgs)) {
		dcc_error_msg("%s", dcc_emsg);
		return 0;
	}
	BUFCPY(info_map_nm, arg);
	set_which_map(MAP_INFO);
	if (!quiet)
		printf("    created %s\n", dcc_fnm2abs_msg(tmp, info_map_nm));
	return 1;
}



static int
info_work(const char *arg, int fgs)
{
	DCC_CLNT_INFO info;
	u_char dcc, srcbad, names;

	if (*arg == '\0') {
		names = 0;
	} else if (!strcmp(arg, "-N")) {
		names = 1;
	} else {
		return -1;
	}

	if (!rdy_ctxt(fgs | DCC_CLNT_FG_BAD_SRVR_OK))
		return 0;

	/* Snapshot the data and then release it while we print it. */
	dcc_ctxts_lock();
	if (!dcc_info_lock(0)) {
		dcc_ctxts_lock();
		return 0;
	}
	memcpy(&info, dcc_clnt_info, sizeof(info));
	srcbad = ctxt && (ctxt->flags & DCC_CTXT_SRCBAD);
	cdcc_unlock(1);

	dcc_rel_priv();
	if (which_map == MAP_INFO) {
		if (info.dcc.nms[0].hostname[0] != '\0'
		    || !grey_on) {
			dcc_print_info(info_map_nm, &info,
				       quiet, 0, srcbad, names,
				       0 <= access(info_map_nm, R_OK));
			dcc = 1;
		} else {
			dcc = 0;
		}
		if (info.grey.nms[0].hostname[0] != '\0'
		    || grey_on) {
			if (dcc && !quiet)
				fputs("\n################\n", stdout);
			dcc_print_info(info_map_nm, &info,
				       quiet, 1, srcbad, names,
				       0 <= access(info_map_nm, R_OK));
		}
	} else {
		dcc_print_info(0, &info, quiet, grey_on ? 2 : 0,
			       srcbad, names, 1);
	}
	if (!quiet)
		putchar('\n');
	return 1;
}



/* server host name */
static int
host_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_SRVR_NM nm;

	if (arg[0] == '\0') {
		if (which_map == MAP_INFO)
			return info_work(arg, DCC_CLNT_FG_BAD_SRVR_OK) ;

		printf("    %s server name \"%s\"\n",
		       grey_on ? "greylist" : "DCC", srvr.hostname);
		return 1;
	}
	if (!strcmp(arg, "-")) {
		set_which_map(MAP_INFO);
		if (!init_map(1, 0)) {
			set_which_map(MAP_TMP);
			return 0;
		}
		return 1;
	}

	arg = dcc_parse_nm_port(0, arg, 0,
				nm.hostname, sizeof(nm.hostname),
				&nm.port, 0, 0,
				0, 0);
	if (!arg)
		return 0;
	arg += strspn(arg, DCC_WHITESPACE);
	if (*arg != '\0')
		return 0;

	set_which_map(MAP_TMP);
	memcpy(srvr.hostname, nm.hostname, sizeof(srvr.hostname));
	if (nm.port != 0) {
		srvr.port = nm.port;
		port_set = 1;
	}

	/* Check that the host name is defined. */
	return rdy_ctxt(DCC_CLNT_FG_BAD_SRVR_OK
			| DCC_CLNT_FG_NO_MEASURE_RTTS);
}



/* server port # */
static int
port_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	u_int port;

	if (arg[0] == '\0') {
		if (which_map == MAP_INFO)
			return info_work(arg, DCC_CLNT_FG_BAD_SRVR_OK) ;
		printf("    port=%d\n", ntohs(srvr.port));
		return 1;
	}

	port = dcc_get_port(0, arg, DCC_GREY2PORT(grey_on), 0, 0);
	if (port == DCC_GET_PORT_INVALID)
		return 0;

	srvr.port = port;
	port_set = 1;
	set_which_map(MAP_TMP);
	return 1;
}



static int
ipv6_cmd(const char *arg, const CMD_TBL_ENTRY *ce)
{
	DCC_INFO_FGS new_info_fgs;
	char sustr[DCC_SU2STR_SIZE];
	sa_family_t check_family;

	if (!init_map(1, 0))
		return 0;

	if (arg[0] == '\0') {
		printf("    %s\n",
		       (info_fgs & DCC_INFO_FG_IPV6_OFF)
		       ? DCC_INFO_TXT_IPV6_OFF
		       : (info_fgs & DCC_INFO_FG_IPV4_OFF)
		       ? DCC_INFO_TXT_IPV6_ONLY
		       : DCC_INFO_TXT_IPV6_ON);
		return 1;
	}

	if (!strcasecmp(arg, "on")) {
		/* "ON" means both IPv4 and IPv6 */
		new_info_fgs = 0;
		check_family = AF_INET6;
	} else if (!strcasecmp(arg, "off")) {
		new_info_fgs = DCC_INFO_FG_IPV6_OFF;
		check_family = AF_INET;
	} else if (!strcasecmp(arg, "only")) {
		new_info_fgs = DCC_INFO_FG_IPV4_OFF;
		check_family = AF_INET6;
	} else {
		return -1;
	}

	if (!ipv_check(dcc_emsg, check_family, 1)) {
		if (new_info_fgs == DCC_INFO_FG_IPV4_OFF) {
			dcc_error_msg("%s", dcc_emsg);
			return 0;
		}
		dcc_error_msg("warning %s", dcc_emsg);
	}

	if (!ck_cmd_priv(ce, 0, 1))
		return 0;

	if (!init_map(1, 1))
		return 0;

	if (((new_info_fgs & DCC_INFO_FG_IPV4_OFF)
	     && dcc_clnt_info->src.family == AF_INET)
	    || ((new_info_fgs & DCC_INFO_FG_IPV6_OFF)
		&& dcc_clnt_info->src.family == AF_INET6)) {
		dcc_error_msg("IPv6 %s incompatible with "
			      DCC_INFO_TXT_USE_SRC"%s"
			      "\n    Remove the source address with 'src -'",
			      arg, dcc_ip2str(sustr, sizeof(sustr),
					      &dcc_clnt_info->src));
		cdcc_unlock(1);
		return 0;
	}

	new_info_fgs |= (info_fgs & ~(DCC_INFO_FG_IPV6_OFF
				      | DCC_INFO_FG_IPV4_OFF));
	if (info_fgs != new_info_fgs) {
		dcc_clnt_info->fgs = info_fgs = new_info_fgs;
		fix_info();
	} else if (!cdcc_unlock(1)) {
		return 0;
	}

	if (rdy_ctxt(0)
	    && 0 != ((new_info_fgs ^ dcc_clnt_info->fgs)
		     & (DCC_INFO_FG_IPV6_OFF | DCC_INFO_FG_IPV4_OFF))) {
#ifdef DCC_NO_IPV6
		dcc_error_msg("IPv6 switch not changed;"
			      " No IPv6 support in this system?");
#else
		dcc_error_msg("IPv6 switch not changed.");
#endif
		return 0;
	}

	return 1;
}



static u_char
ck_new_src(DCC_EMSG emsg, DCC_IP *new_ip, const char *arg, DCC_INFO_FGS fgs)
{
	DCC_SOCKET soc;
	DCC_SOCKU su;
	int error;

	if (!strcmp(arg, "-")) {
		memset(new_ip, 0, sizeof(*new_ip));
		return 1;
	}

	if (!str2su(&su, arg)) {
		/* If our source address is IPv4 (or IPv6), then
		 * we can use only IPv4 (or IPv6) DCC servers */
		dcc_host_lock();
		if (!dcc_get_host(arg,
				  (fgs & DCC_INFO_FG_IPV6_OFF) ? GETH_V4
				  : (fgs & DCC_INFO_FG_IPV4_OFF) ? GETH_V6
				  : DCC_GETH_DEF,
				  &error)) {
			dcc_host_unlock();
			dcc_pemsg(EX_NOHOST, emsg, "%s: %s",
				  arg, H_ERROR_STR1(error));
			return 0;
		}
		su = dcc_hostaddrs[0];
		dcc_host_unlock();

	} else if (su.sa.sa_family == AF_INET) {
		if (fgs & DCC_INFO_FG_IPV4_OFF) {
			dcc_pemsg(EX_NOHOST, emsg,
				  "IPv4 source address"
				  " not possible when IPv4 disabled");
			return 0;
		}
	} else {
		if (fgs & DCC_INFO_FG_IPV6_OFF) {
			dcc_pemsg(EX_NOHOST, emsg,
				  "IPv6 source address"
				  " not possible when IPv6 disabled");
			return 0;
		}
	}

	soc = INVALID_SOCKET;
	if (!udp_create(dcc_emsg, &soc, &su, 0))
		return 0;
	closesocket(soc);

	dcc_su2ip(new_ip, &su);
	return 1;
}



static int
src_cmd(const char *arg, const CMD_TBL_ENTRY *ce)
{
	DCC_IP new_ip;
	char sustr[DCC_SU2STR_SIZE];
	int result;

	if (arg[0] == '\0') {
		if (!init_map(1, 0))
			return 0;

		if (dcc_clnt_info->src.family == AF_UNSPEC) {
			printf("    no source address specified\n");
		} else {
			/* display what the system actually uses */
			printf("    source address=%s%s\n",
			       dcc_su2str2(sustr, sizeof(sustr),
					   &ctxt->soc[0].loc),
			       (ctxt->flags & DCC_CTXT_SRCBAD)
			       ? " "DCC_INFO_TXT_USE_SRCBAD : "");
		}
		return 1;
	}

	if (!ck_cmd_priv(ce, 0, 1))
		return 0;

	if (!init_map(1, 1))
		return 0;

	/* check the host name after we have locked the IPv4 vs. IPv6 choice */
	if (!ck_new_src(dcc_emsg, &new_ip, arg, dcc_clnt_info->fgs)) {
		dcc_error_msg("%s", dcc_emsg);
		result = 0;
	} else {
		result = 1;
		src = new_ip;
		dcc_clnt_info->src = src;
	}
	fix_info();
	return result;
}



static int
socks_cmd(const char *arg, const CMD_TBL_ENTRY *ce)
{
	DCC_INFO_FGS new_use_socks;

	if (arg[0] == '\0') {
		if (!init_map(1, 0))
			return 0;
		printf("    SOCKS %s\n",
		       (info_fgs & DCC_INFO_FG_SOCKS) ? "on" : "off");
		return 1;
	}

	if (!strcmp(arg, "off")) {
		new_use_socks = 0;
	} else if (!strcmp(arg, "on")) {
		new_use_socks = DCC_INFO_FG_SOCKS;
	} else {
		return -1;
	}

	if (!ck_cmd_priv(ce, 0, 1))
		return 0;

	if (!init_map(1, 1))
		return 0;
	if ((dcc_clnt_info->fgs & DCC_INFO_FG_SOCKS) == new_use_socks)
		return cdcc_unlock(1);	/* nothing to do */

	dcc_clnt_info->fgs ^= DCC_INFO_FG_SOCKS;
	info_fgs = dcc_clnt_info->fgs;

	fix_info();
	return 1;
}



static int
passwd_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_PASSWD new_passwd;
	DCC_PATH tmp;

	if (arg[0] == '\0') {
		if (which_map == MAP_INFO) {
			printf("    using password in %s\n",
			       dcc_fnm2abs_msg(tmp, info_map_nm));
			if (passwd_set)
				printf("    but the password for explicitly"
				       " named servers is "DCC_PASSWD_PAT"\n",
				       passwd);
		} else {
			if (passwd_set) {
				printf("    password "DCC_PASSWD_PAT"\n",
				       passwd);
			} else if (srvr.passwd[0] != '\0') {
				printf("    password not set;"
				       " using "DCC_PASSWD_PAT" from %s\n",
				       srvr.passwd,
				       dcc_fnm2abs_msg(tmp, ids_path));
			} else {
				printf("    password not set\n");
			}
		}
		return 1;
	}

	arg = parse_passwd(0, new_passwd, arg, "password", 0, 0);
	if (!arg || *arg != '\0')
		return -1;
	memcpy(passwd, new_passwd, sizeof(passwd));
	passwd_set = 1;
	set_which_map(MAP_TMP);
	return 1;
}



static int
id_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_CLNT_ID id;

	if (arg[0] == '\0') {
		printf("    ID=%d\n", srvr_clnt_id);
		return 1;
	}

	id = dcc_get_id(0, arg, 0, 0);
	if (id == DCC_ID_INVALID)
		return -1;

	srvr_clnt_id = id;
	set_which_map(MAP_TMP);
	return 1;
}



static int
debug_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	char debug_str[24];
	char ttl_str[24];
	u_long new_ttl, l, new_debug;
	char *p;

	if (arg[0] == '\0') {
		if (!dcc_clnt_debug)
			snprintf(debug_str, sizeof(debug_str),
				 "debug off");
		else if (dcc_clnt_debug == 1)
			snprintf(debug_str, sizeof(debug_str),
				 "debug on");
		else
			snprintf(debug_str, sizeof(debug_str),
				 "debug on+%d\n", dcc_clnt_debug-1);
		if (dcc_debug_ttl != 0)
			snprintf(ttl_str, sizeof(ttl_str),
				 "    TTL=%d", dcc_debug_ttl);
		else
			ttl_str[0] = '\0';
		printf("    %s%s\n", debug_str, ttl_str);
		return 1;
	}

	new_ttl = dcc_debug_ttl;
	new_debug = dcc_clnt_debug;
	for (;;) {
		if (!CLITCMP(arg, "off")) {
			new_debug = 0;
			arg += LITZ("off");
		} else if (!CLITCMP(arg, "on")) {
			++new_debug;
			arg += LITZ("on");
		} else if (!CLITCMP(arg, "ttl=")) {
			new_ttl = strtoul(arg+LITZ("ttl="), &p, 10);
#if defined(IPPROTO_IP) && defined(IP_TTL)
			if (new_ttl < 256)
				arg = p;
#else
			printf("    TTL setting not supported\n");
#endif
		} else if ((l = strtoul(arg, &p, 10)) > 0 && l < 10) {
			new_debug = l;
			arg = p;
		}

		if (*arg == '\0')
			break;
		if (*arg == ' ' || *arg == '\t') {
			arg += strspn(arg, DCC_WHITESPACE);
		} else {
			return -1;
		}
	}

	dcc_debug_ttl = new_ttl;
	if (dcc_debug_ttl != 0)
		set_which_map(MAP_TMP);

	dcc_clnt_debug = new_debug;
	if (dcc_clnt_debug > 1)
		printf("    debug on+%d\n", dcc_clnt_debug-1);
	return 1;
}



static int
no_fail_cmd(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (!init_map(1, 1))
		return 0;
	DCC_GREY2CLASS(grey_on)->fail_time = 0;
	cdcc_unlock(1);
	return 1;
}



static int
quiet_cmd(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (arg[0] == '\0') {
		printf("    %s\n", quiet ? "on" : "off");
		return 1;
	} else if (!CLITCMP(arg, "on")) {
		quiet = 1;
		return 1;
	} else if (!CLITCMP(arg, "off")) {
		quiet = 0;
		return 1;
	}
	return -1;
}



static int
ckludge_cmd(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	char *p;
	long l;

	if (arg[0] == '\0') {
		printf("    clock kludge=%d\n", (int)clock_kludge);
		return 1;
	}

	l = strtol(arg, &p, 10);
	if (*p != '\0') {
		dcc_error_msg("invalid clock kludge \"%s\"", arg);
		return -1;
	}
	clock_kludge = l;
	return 1;
}



/* parse a line of the following form
 *	name[,port-#] [RTT+adj] [Greylist] [client-ID [password]]
 *  The port-# can be "-" to specifiy the default DCC server port.
 *  If both the client-ID and the password are absent, then the anonymous
 *  client-ID is used.
 *  A null string is assumed if the password is missing.
 */
static int				/* 0=minor, -1=can't add, -2=can't del */
parse_srvr_nm(DCC_EMSG emsg,
		  DCC_SRVR_NM *nmp,	/* build this entry */
		  u_char *pgrey,	/* 1=for greylisting */
		  const char *line,	/* from this string */
		  const char *fnm, int lno) /* that came from here */
{
	char id_buf[12];
	char port_buf[3];
	int result;
	char *p;
	long l;

	result = 1;
	memset(nmp, 0, sizeof(DCC_SRVR_NM));

	line = dcc_parse_nm_port(emsg, line, DCC_GREY2PORT(pgrey && *pgrey),
				 nmp->hostname, sizeof(nmp->hostname),
				 &nmp->port, port_buf, sizeof(port_buf),
				 fnm, lno);
	if (!line)
		return -2;

	/* look for RTT and greylist flag in either order */
	for (;;) {
		/* look for greylist flag */
		if (!CLITCMP(line, "greylist")
		    && (line[LITZ("greylist")] == '\0'
			|| line[LITZ("greylist")] == ' '
			|| line[LITZ("greylist")] == '\t')) {
			line += LITZ("greylist")+strspn(line+LITZ("greylist"),
							DCC_WHITESPACE);
			if (pgrey)
				*pgrey = 1;
			if (port_buf[0] == '\0' || !strcmp(port_buf, "-"))
				nmp->port = htons(DCC_GREY_PORT);
			continue;
		}

		/* look for optional RTT adjustment */
		if (CLITCMP(line, "rtt"))
			break;
		line += LITZ("rtt")+strspn(line+LITZ("rtt"), DCC_WHITESPACE);
		l = strtol(line, &p, 10);
		if (p != line) {
			int wsp = strspn(p, DCC_WHITESPACE);
			if (!CLITCMP(p+wsp, "ms"))
				p += wsp+LITZ("ms");
		}
		if (p == line
		    || (*p != '\0' && *p != ' ' && *p != '\t')) {
			dcc_pemsg(EX_DATAERR, emsg,
				  "invalid RTT adjustment%s",
				  dcc_fnm_lno(&fnm_buf, fnm, lno));
			result = 0;
		} else {
			if (l < -DCC_RTT_ADJ_MAX/1000) {
				l = -DCC_RTT_ADJ_MAX/1000;
				dcc_pemsg(EX_DATAERR, emsg,
					  "invalid RTT adjustment%s",
					  dcc_fnm_lno(&fnm_buf, fnm, lno));
				result = 0;
			} else if (l > DCC_RTT_ADJ_MAX/1000) {
				l = DCC_RTT_ADJ_MAX/1000;
				dcc_pemsg(EX_DATAERR, emsg,
					  "invalid RTT adjustment%s",
					  dcc_fnm_lno(&fnm_buf, fnm, lno));
				result = 0;
			}
			nmp->rtt_adj = l*1000;
			line = p+strspn(p, DCC_WHITESPACE);
		}
	}

	/* get the client-ID */
	line = dcc_parse_word(emsg, id_buf, sizeof(id_buf),
			      line, "client-ID", fnm, lno);
	if (!line)
		return -2;
	if (id_buf[0] == '\0') {
		nmp->clnt_id = DCC_ID_ANON;
	} else {
		nmp->clnt_id = dcc_get_id(emsg, id_buf, fnm, lno);
		if (nmp->clnt_id == DCC_ID_INVALID)
			return -1;
		if (nmp->clnt_id < DCC_CLNT_ID_MIN
		    && nmp->clnt_id != DCC_ID_ANON) {
			dcc_pemsg(EX_DATAERR, emsg,
				  "server-ID %d is not a client-ID",
				  nmp->clnt_id);
			return -1;
		}
	}

	/* allow null password only for anonymous clients */
	if (*line == '\0') {
		if (nmp->clnt_id != DCC_ID_ANON) {
			dcc_pemsg(EX_DATAERR, emsg,
				  "invalid null password for client-ID %d"
				  " for %s%s",
				  nmp->clnt_id, nmp->hostname,
				  dcc_fnm_lno(&fnm_buf, fnm, lno));
			return -1;
		}
		return result;
	}

	line = parse_passwd(emsg, nmp->passwd, line, "passwd", fnm, lno);
	if (nmp->passwd[0] == '\0' || *line != '\0') {
		dcc_pemsg(EX_DATAERR, emsg,
			  "invalid password server %s, client-ID %d%s",
			  nmp->hostname, nmp->clnt_id,
			  dcc_fnm_lno(&fnm_buf, fnm, lno));
		if (nmp->clnt_id != DCC_ID_ANON)
			return -1;
		result = 0;
	}

	if (nmp->clnt_id == DCC_ID_ANON) {
		dcc_pemsg(EX_DATAERR, emsg,
			  "password invalid for %s with anonymous client-ID%s",
			  nmp->hostname,
			  dcc_fnm_lno(&fnm_buf, fnm, lno));
		result = 0;
	}

	return result;
}



static int
delete_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_SRVR_CLASS *class;
	DCC_SRVR_NM nm, *nmp;
	DCC_SRVR_ADDR *addr;
	u_char del_grey;

	del_grey = grey_on;
	if (-1 > parse_srvr_nm(dcc_emsg, &nm, &del_grey, arg, 0, 0)) {
		dcc_error_msg("%s", dcc_emsg);
		return 0;
	}

	/* map and lock */
	set_which_map(MAP_INFO);
	if (!init_map(1, 1))
		return 0;

	class = DCC_GREY2CLASS(del_grey);
	for (nmp = class->nms; nmp <= LAST(class->nms); ++nmp) {
		if (strcasecmp(nmp->hostname, nm.hostname)
		    || nmp->port != nm.port)
			continue;

		/* Found it. */

		/* zap its IP addresses so they won't be used
		 * if resolving the remaining names fails */
		for (addr = class->addrs; addr <= LAST(class->addrs); ++addr) {
			if (addr->nam_inx == nmp - class->nms) {
				addr->rtt = DCC_RTT_BAD;
				addr->nam_inx = NO_NAM;
			}
		}
		if (nmp != LAST(class->nms))
			memmove(nmp, nmp+1,
				(LAST(class->nms) - nmp)*sizeof(*nmp));
		memset(LAST(class->nms), 0, sizeof(*nmp));
		++class->gen;
		fix_info();
		return 1;
	}

	dcc_error_msg("server \"%s,%d\" not found",
		      nm.hostname, ntohs(nm.port));
	cdcc_unlock(1);
	return 0;
}



static int
add_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_SRVR_CLASS *class;
	DCC_SRVR_NM nm, *nmp, *tgt_nmp;
	u_char add_grey;
	int i;

	add_grey = grey_set && grey_on;

	i = parse_srvr_nm(dcc_emsg, &nm, &add_grey, arg, 0, 0);
	if (i < 0) {
		dcc_error_msg("%s", dcc_emsg);
		return 0;
	}
	if (nm.clnt_id == DCC_ID_ANON && add_grey) {
		dcc_error_msg("anonymous client-ID invalid"
			      " for Greylist server %s",
			      nm.hostname);
		return 0;
	}
	if (i < 1)
		dcc_error_msg("    %s", dcc_emsg);

	/* map and lock the information */
	set_which_map(MAP_INFO);
	if (!init_map(1, 1))
		return 0;

	/* look for the old entry or a new, free entry */
	class = DCC_GREY2CLASS(add_grey);
	tgt_nmp = 0;
	for (nmp = class->nms; nmp <= LAST(class->nms); ++nmp) {
		if (nmp->hostname[0] == '\0') {
			if (!tgt_nmp)
				tgt_nmp = nmp;
			continue;
		}
		if (!strcmp(nmp->hostname, nm.hostname)
		    && nmp->port == nm.port) {
			printf("    overwriting existing entry\n");
			tgt_nmp = nmp;
			break;
		}
	}

	if (tgt_nmp) {
		memcpy(tgt_nmp, &nm, sizeof(*tgt_nmp));
		fix_info();
		return 1;
	}

	cdcc_unlock(1);
	if (add_grey)
		dcc_error_msg("too many Greylist server names");
	else
		dcc_error_msg("too many DCC server names");
	return 0;
}



static void
add_new_nms(const DCC_SRVR_NM new_nms[DCC_MAX_SRVR_NMS],
	    DCC_SRVR_NM old_nms[DCC_MAX_SRVR_NMS])
{
	const DCC_SRVR_NM *new_nmp;
	DCC_SRVR_NM *old_nmp;

	for (new_nmp = new_nms;
	     new_nmp < &new_nms[DCC_MAX_SRVR_NMS]
	     && new_nmp->hostname[0] != '\0';
	     ++new_nmp) {
		for (old_nmp = old_nms;
		     old_nmp <= &old_nms[DCC_MAX_SRVR_NMS];
		     ++old_nmp) {
			if (old_nmp->hostname[0] == '\0'
			    || (!strcmp(old_nmp->hostname, new_nmp->hostname)
				&& old_nmp->port == new_nmp->port)) {
				memcpy(old_nmp, new_nmp, sizeof(*old_nmp));
				break;
			}
		}
	}
}



/* open a file of server names, ports, and so forth */
FILE *
open_map_txt(DCC_EMSG emsg, const char *nm)
{
	FILE *f;
	DCC_PATH path;

	f = fopen(nm, "r");
	if (!f) {
		dcc_pemsg(EX_NOINPUT, emsg, "fopen(%s): %s",
			  dcc_fnm2abs_msg(path, nm), ERROR_STR());
		return 0;
	}

	/* the file contains passwords,
	 * so refuse to use it if everyone can read it */
	if (!dcc_ck_private(emsg, 0, nm, fileno(f))) {
		fclose(f);
		return 0;
	}
	return f;
}



static int				/* -1=bad line 0=not option line 1=ok */
get_options(DCC_EMSG emsg, const char *cp,
	    DCC_INFO_FGS *new_info_fgs, DCC_IP *new_src)
{
	u_char option_line;
	char src_addr[INET6_ADDRSTRLEN+1];
	sa_family_t check_family;
	char *p;
	int version;

	option_line = 0;
	version = 0;
	src_addr[0] = '\0';
	for (;;) {
		cp += strspn(cp, DCC_WHITESPACE);
		if (*cp == '\0')
			break;

		if (!CLITCMP(cp, DCC_INFO_TXT_IPV6)) {
			if (!CLITCMP(cp, DCC_INFO_TXT_IPV6_OFF)) {
				cp += LITZ(DCC_INFO_TXT_IPV6_OFF);
				*new_info_fgs &= ~DCC_INFO_FG_IPV4_OFF;
				*new_info_fgs |= DCC_INFO_FG_IPV6_OFF;
				check_family = AF_INET;

			} else if (!CLITCMP(cp, DCC_INFO_TXT_IPV6_ON)) {
				cp += LITZ(DCC_INFO_TXT_IPV6_ON);
				*new_info_fgs &= ~(DCC_INFO_FG_IPV4_OFF
						   | DCC_INFO_FG_IPV6_OFF);
				check_family = AF_INET6;

			} else if (!CLITCMP(cp, DCC_INFO_TXT_IPV6_ONLY)) {
				cp += LITZ(DCC_INFO_TXT_IPV6_ONLY);
				*new_info_fgs &= ~DCC_INFO_FG_IPV6_OFF;
				*new_info_fgs |= DCC_INFO_FG_IPV4_OFF;
				check_family = AF_INET6;
			} else {
				break;
			}

			if (dcc_clnt_debug
			    && !ipv_check(dcc_emsg, check_family, 1))
				dcc_error_msg("warning %s", dcc_emsg);

			option_line = 1;
			continue;
		}

		if (!CLITCMP(cp, DCC_INFO_TXT_USE_SOCKS)) {
			cp += LITZ(DCC_INFO_TXT_USE_SOCKS);
			cp += strspn(cp, DCC_WHITESPACE);
			*new_info_fgs |= DCC_INFO_FG_SOCKS;
			option_line = 1;
			continue;
		}

		if (!CLITCMP(cp, DCC_INFO_TXT_USE_SRC)) {
			cp += LITZ(DCC_INFO_TXT_USE_SRC);
			cp = dcc_parse_word(emsg, src_addr, sizeof(src_addr),
					    cp, 0, 0, 0);
			if (!cp)
				return -1;
			/* ignore old complaints about the source */
			if (!CLITCMP(cp, DCC_INFO_TXT_USE_SRCBAD)) {
				cp += LITZ(DCC_INFO_TXT_USE_SRCBAD);
				cp += strspn(cp, DCC_WHITESPACE);
			}
			option_line = 1;
			continue;
		}

		if (!CLITCMP(cp, DCC_INFO_TXT_VERSION)) {
			cp += LITZ(DCC_INFO_TXT_VERSION);
			version = strtol(cp, &p, 10);
			if (version != 3
			    || (*p != ' ' && *p != '\t'
				&& *p != '\n' && *p != '\r')){
				dcc_pemsg(EX_DATAERR, emsg,
					  "unrecognized version number");
				return -1;
			}
			cp = p;
			option_line = 1;
			continue;
		}

		if (option_line)
			break;
		return 0;
	}

	if (option_line && *cp != '\0') {
		dcc_pemsg(EX_DATAERR, emsg, "unrecognized option \"%s\"", cp);
		return -1;
	}

	/* ignore IPv6 settings in old files */
	if (version < 3)
		*new_info_fgs &= ~(DCC_INFO_FG_IPV4_OFF | DCC_INFO_FG_IPV6_OFF);

	if (src_addr[0] != '\0'
	    && !ck_new_src(dcc_emsg, new_src, src_addr, *new_info_fgs))
		return -1;

	return 1;
}



static int
load_cmd(const char *lfile, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_INFO_FGS new_info_fgs;
	u_char load_grey;
	DCC_SRVR_NM new_nm;
	DCC_SRVR_NM dcc_nms[DCC_MAX_SRVR_NMS];
	int num_dcc_nms;
	DCC_SRVR_NM grey_nms[DCC_MAX_SRVR_NMS];
	int num_grey_nms;
	int got_options;
	DCC_IP new_src;
	char buf[sizeof(DCC_SRVR_NM)*3];
	const char *bufp;
	DCC_PATH tmp;
	FILE *f;
	int fd, lno;
	int i;

	if (*lfile == '\0')
		return -1;

	dcc_rel_priv();
	if (!strcmp(lfile,"-")) {
		lfile = 0;
		fd = dup(fileno(stdin));
		if (fd < 0) {
			dcc_error_msg("dup(stdin): %s", ERROR_STR());
			return 0;
		}
		f = fdopen(fd, "r");
		if (!f) {
			dcc_error_msg("fdopen(): %s", ERROR_STR());
			return 0;
		}
	} else {
		f = open_map_txt(dcc_emsg, lfile);
		if (!f) {
			dcc_error_msg("%s", dcc_emsg);
			return 0;
		}
	}

	/* parse the text file to create a pair of lists of server names */
	new_info_fgs = info_fgs;
	got_options = -1;
	num_dcc_nms = 0;
	memset(dcc_nms, 0, sizeof(dcc_nms));
	num_grey_nms = 0;
	memset(grey_nms, 0, sizeof(grey_nms));
	memset(&new_src, 0, sizeof(new_src));
	lno = 0;
	for (;;) {
		bufp = fgets(buf, sizeof(buf), f);
		if (!bufp) {
			if (ferror(f)) {
				dcc_error_msg("fgets(%s): %s",
					      !lfile
					      ? "STDIN"
					      : dcc_fnm2abs_msg(tmp, lfile),
					      ERROR_STR());
				fclose(f);
				return 0;
			}
			break;
		}

		++lno;

		/* skip blank lines and comments */
		bufp += strspn(bufp, DCC_WHITESPACE);
		if (*bufp == '\0' || *bufp == '#')
			continue;

		/* look for version # and flags in the first non-comment line */
		if (got_options < 0) {
			got_options = get_options(dcc_emsg,
						  bufp, &new_info_fgs, &new_src);
			if (got_options < 0) {
				dcc_error_msg("%s%s", dcc_emsg,
					      dcc_fnm_lno(&fnm_buf, lfile, lno));
				fclose(f);
				return 0;
			}
			if (got_options > 0)
				continue;
		}

		load_grey = 0;
		i = parse_srvr_nm(dcc_emsg, &new_nm, &load_grey,
				  bufp, lfile, lno);
		if (i < 0) {
			dcc_error_msg("%s", dcc_emsg);
			fclose(f);
			return 0;
		}
		if (i < 1)
			dcc_error_msg("    %s", dcc_emsg);
		if (load_grey) {
			if (new_nm.clnt_id == DCC_ID_ANON) {
				dcc_error_msg("anonymous client-ID invalid"
					      " for Greylist server %s%s",
					      new_nm.hostname,
					      dcc_fnm_lno(&fnm_buf, lfile, lno));
				fclose(f);
				return 0;
			}
			if (num_grey_nms >= DIM(grey_nms)) {
				dcc_error_msg("too many Greylist server names"
					      "%s",
					      dcc_fnm_lno(&fnm_buf, lfile, lno));
				fclose(f);
				return 0;
			}
			grey_nms[num_grey_nms++] = new_nm;
		} else {
			if (num_dcc_nms >= DIM(dcc_nms)) {
				dcc_error_msg("too many DCC server names%s",
					      dcc_fnm_lno(&fnm_buf, lfile, lno));
				fclose(f);
				return 0;
			}
			dcc_nms[num_dcc_nms++] = new_nm;
		}
	}
	fclose(f);
	if (!got_options && num_grey_nms == 0 && num_dcc_nms == 0) {
		dcc_error_msg("no DCC server names or options in %s",
			      dcc_fnm2abs_msg(tmp, lfile));
		return 0;
	}

	/* create the map, without set-UID powers to prevent games,
	 * and then lock, install, and unlock the information */
	dcc_rel_priv();

	if (which_map != MAP_INFO)
		set_which_map(MAP_INFO);
	if (!init_map(0, 1)) {
		/* create a new map */
		if (!dcc_create_map(0, info_map_nm, 0,
				    0, 0, 0, 0, &new_src, new_info_fgs))
			return 0;
		printf("    created %s\n", dcc_fnm2abs_msg(tmp, info_map_nm));
		if (!init_map(1, 1))
			return 0;
	}

	/* merge the old and new entries */
	add_new_nms(grey_nms, dcc_clnt_info->grey.nms);
	add_new_nms(dcc_nms, dcc_clnt_info->dcc.nms);
	dcc_clnt_info->fgs = info_fgs = new_info_fgs;
	if (new_src.family != AF_UNSPEC)
		dcc_clnt_info->src = new_src;

	fix_info();

	if (!quiet) {
		if (!lfile)
			printf("##################\n\n");
		return info_work("", DCC_CLNT_FG_BAD_SRVR_OK) ;
	}
	return 1;
}



static int
info_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	/* map, copy, and unlock the information
	 * prefer to talk to the server, but don't wait
	 * unless we have changed the file and are not being quiet */
	return info_work(arg,
			 (map_changed && !quiet)
			 ? DCC_CLNT_FG_BAD_SRVR_OK
			 : (DCC_CLNT_FG_NO_MEASURE_RTTS
			    | DCC_CLNT_FG_BAD_SRVR_OK));
}



static int
rtt_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (!init_map(1, 1))
		return 0;
	dcc_force_measure_rtt(&dcc_clnt_info->dcc);
	dcc_force_measure_rtt(&dcc_clnt_info->grey);
	cdcc_unlock(1);

	/* wait to talk to the server, but don't insist */
	return info_work(arg, quiet ? DCC_CLNT_FG_BAD_SRVR_OK : 0);
}


/* delete a checksum */
static int				/* 1=ok, 0=bad checksum, -1=fatal */
delck_sub(DCC_EMSG emsg, WF *wf DCC_UNUSED,
	  DCC_CK_TYPES type, DCC_SUM *sum, DCC_TGTS tgts DCC_UNUSED)
{
	struct timeval cmd_start, cmd_end;
	char type_buf[DCC_XHDR_MAX_TYPE_LEN_BUF];
	char ck_buf[sizeof(DCC_SUM)*3+2];
	DCC_DELETE del;
	DCC_OP_RESP resp;
	char ob[DCC_OPBUF];
	u_char result;

	printf(" deleting %s  %s\n",
	       type2str(type_buf, sizeof(type_buf), type, 0, 1, grey_on),
	       ck2str(ck_buf, sizeof(ck_buf), type, sum, 0));

	memset(&del, 0, sizeof(del));
	gettimeofday(&cmd_start, 0);
	del.date = htonl(cmd_start.tv_sec);
	del.ck.type = type;
	del.ck.len = sizeof(del.ck);
	memcpy(&del.ck.sum, sum, sizeof(DCC_SUM));
	result = dcc_clnt_op(emsg, ctxt, grey_clnt_fgs | DCC_CLNT_FG_NO_FAIL,
			     0, 0, 0, &del.hdr, sizeof(del),
			     DCC_OP_DELETE, &resp, sizeof(resp));
	gettimeofday(&cmd_end, 0);
	if (!result) {
		dcc_error_msg("%s", dcc_emsg);
	} else {
		switch (resp.hdr.op) {
		case DCC_OP_OK:
			break;

		case DCC_OP_ERROR:
			cdcc_error_msg("   %.*s",
				       (ntohs(resp.hdr.len)
					-(int)(sizeof(resp.error)
					       - sizeof(resp.error.msg))),
				       resp.error.msg);
			result = 0;
			break;

		default:
			dcc_error_msg("unexpected response: %s",
				      dcc_hdr_op2str(ob,sizeof(ob), &resp.hdr));
			result = 0;
			break;
		}
	}

	if (dcc_clnt_debug) {
		printf("%.2f ms\n",
		       ((cmd_end.tv_sec-cmd_start.tv_sec)*1000.0
			+ (cmd_end.tv_usec-cmd_start.tv_usec)/1000.0));
	}
	return result;
}



/* delete a simple checksum */
static int
delck_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	char type_str[DCC_XHDR_MAX_TYPE_LEN_BUF];

	if (*arg == '\0')
		return -1;
	arg = dcc_parse_word(0, type_str, sizeof(type_str),
			     arg, 0, 0, 0);
	if (!arg)
		return -1;

	if (!rdy_ctxt(0))
		return 0;
	return 0 < dcc_parse_hex_ck(0, &cmn_wf,
				    type_str, dcc_str2type_del(type_str, -1),
				    arg, 0, delck_sub);
}



static int
sleep_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	double s;
	char *p;

	s = strtod(arg, &p);
	if (*p != '\0' || s < 0.001 || s > 1000)
		return -1;
	usleep((u_int)(s*1000000.0));
	return 1;
}




static const u_char *
client_unpack64(const u_char *cp, u_int64_t *vp)
{
	u_char c;
	u_int64_t v;
	int shift;

	v = 0;
	shift = 0;
	do {
		c = *cp++;
		v |= (c & 0x7f) << shift;
		shift += 7;
	} while (c & 0x80);

	*vp = v;
	return cp;
}



static int
client_unpack(const u_char *cp0,
	      u_char *flagsp,
	      DCC_CLNT_ID *clnt_idp,
	      u_int *offsetp,
	      DCC_PTIME *last_usedp,
	      DCC_PTIME *prev_last_usedp,
	      DCC_SCNTR *requestsp,
	      DCC_SCNTR *nopsp,
	      u_char *versp,
	      DCC_SOCKU *su)
{
	const u_char *cp;
	u_char flags;
	u_int64_t v;
	struct in6_addr in6_addr;
	struct in_addr in_addr;

#ifdef DCC_PKT_VERSION6
	if (aop_resp.hdr.pkt_vers <= DCC_PKT_VERSION6) {
#define CPY2(s) ((s[0]<<8) | s[1])
#define CPY3(s) ((s[0]<<16) | (s[1]<<8) | s[2])
#define CPY4(s) ((s[0]<<24) | (s[1]<<16) | (s[2]<<8) | s[3])
		const DCC_ADMN_RESP_CLIENTSv6 *cl;

		cl = (DCC_ADMN_RESP_CLIENTSv6 *)cp0;
		flags = cl->flags;
		*flagsp = flags & (DCC_ADMN_RESP_CLIENTS_BL
				   | DCC_ADMN_RESP_CLIENTS_SKIP);
		*clnt_idp = CPY4(cl->clnt_id);
		if (flags & DCC_ADMN_RESP_CLIENTS_SKIP) {
			/* skip place keepers */
			*offsetp += CPY3(cl->requests);
			*last_usedp = 0;
			*requestsp = 0;
		} else {
			*last_usedp = CPY4(cl->last_used);
			*requestsp = CPY3(cl->requests);
		}
		*prev_last_usedp = 0;
		*nopsp = CPY2(cl->nops);
		if (flags & DCC_ADMN_RESP_CLIENTS_IPV6) {
			memcpy(&in6_addr, &cl->addr, sizeof(in6_addr));
			dcc_mk_inet6_su(su, &in6_addr, 0, 0);
			return (sizeof(*cl) - sizeof(cl->addr)
				+ sizeof(cl->addr.ipv6));
		}
		memcpy(&in_addr, &cl->addr, sizeof(in_addr));
		dcc_mk_inet_su(su, &in_addr, 0);
		return (sizeof(*cl) - sizeof(cl->addr)
			+ sizeof(cl->addr.ipv4));
	}
#undef CPY2
#undef CPY3
#undef CPY4
#endif

	cp = cp0;
	flags = *cp++;
	*flagsp = flags & (DCC_ADMN_RESP_CLIENTS_BL
			   | DCC_ADMN_RESP_CLIENTS_BAD
			   | DCC_ADMN_RESP_CLIENTS_SKIP
			   | DCC_ADMN_RESP_CLIENTS_LAST);
	/* if the version is absent,
	 * then it must be the same as the previous value */
	if (flags & DCC_ADMN_RESP_CLIENTS_VERS)
		*versp = *cp++;
#ifdef DCC_PKT_VERSION9
	if (aop_resp.hdr.pkt_vers <= DCC_PKT_VERSION9) {
		v = *cp++ << 24;
		v |= *cp++ << 16;
		v |= *cp++ << 8;
		v |= *cp++;
		*prev_last_usedp = 0;
	} else
#endif
	cp = client_unpack64(cp, &v);
	if (flags & DCC_ADMN_RESP_CLIENTS_SKIP) {
		*offsetp += v;
		*last_usedp = 0;
	} else {
		if (v <= DCC_ADMIN_CLIENTS_MAX_DELTA)
			v += *prev_last_usedp;
		*prev_last_usedp = v;
		*last_usedp = v;
	}

	if ((flags & DCC_ADMN_RESP_CLIENTS_ID1) != 0)
		v = DCC_ID_ANON;
	else
		cp = client_unpack64(cp, &v);
	*clnt_idp = v;

	cp = client_unpack64(cp, requestsp);
	cp = client_unpack64(cp, nopsp);

	if (flags & DCC_ADMN_RESP_CLIENTS_IPV6) {
		memcpy(&in6_addr, cp, sizeof(in6_addr));
		dcc_mk_inet6_su(su, &in6_addr, 0, 0);
		cp += 16;
	} else {
		memcpy(&in_addr, cp, sizeof(in_addr));
		dcc_mk_inet_su(su, &in_addr, 0);
		cp += 4;
	}

	return cp - cp0;
}



/* get the server's list of recent clients */
static int
clients_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	u_char nonames, sort, ids, req_flags;
	u_char passed_flags, passed_max_clients, passed_thold, passed_cidr;
	struct in6_addr addr6;
	DCC_AOP_CLIENTS_CIDR addr_bits;
	u_int max_clients, thold;
	u_int total, subtotal;
	DCC_SCNTR max_ops, max_nops;
	int ops_width, nops_width;
	u_int offset;			/* next client wanted from server */
	u_int num_clients, list_len;
	DCC_SOCKU su;
	struct ct {
	    struct ct	*lt, *gt;	/* sorted order */
	    DCC_PTIME	last_used;
	    DCC_SCNTR	requests;
	    DCC_SCNTR	nops;
	    u_char	flags;
	    u_char	vers;
	    DCC_CLNT_ID	clnt_id;
	    DCC_SOCKU	su;
	} *cttop, *ctbot, *ctnew, *ct;
	DCC_SCNTR versions[DCC_PKT_VERSION_MAX+1];

	char date_buf[40];
	struct tm last, now;
	char *p;
	const char *ac;
	u_char need_head;
	u_int n;
	int i;

	passed_flags = 0;
	thold = 0;
	passed_thold = 0;
	max_clients = DCC_ADMIN_RESP_MAX_CLIENTS;
	passed_max_clients = 0;
	passed_cidr = 0;
	memset(addr_bits, 0, sizeof(addr_bits));

	ac = strpbrk(arg, "/.:");

	/* look for "-n", "-ns", "-n -s", etc. */
	nonames = 0;
	sort = 0;
	ids = 0;
	req_flags = 0;
	while (*arg != 0) {
		arg += strspn(arg, " \t");
		if (*arg == '-' && !passed_flags) {
			++arg;
			do {
				switch (*arg) {
				case 'n':
					nonames = 1;
					break;
				case 's':
					sort = 1;
					break;
				case 'i':
					ids = 1;
					break;
				case 'a':
					req_flags |= DCC_AOP_CLIENTS_AVG;
					break;
				case 'V':
					req_flags |= DCC_AOP_CLIENTS_VERS;
					break;
				case 'A':
					req_flags |= DCC_AOP_CLIENTS_ANON;
					req_flags &= ~DCC_AOP_CLIENTS_NON_ANON;
					break;
				case 'K':
					req_flags |= DCC_AOP_CLIENTS_NON_ANON;
					req_flags &= ~DCC_AOP_CLIENTS_ANON;
					break;
				default:
					help_cmd("clients", 0);
					return -1;
				}
			} while (*++arg != ' ' && *arg != '\t' && *arg != '\0');
			continue;
		}
		if (!passed_cidr && ac && !strpbrk(arg, DCC_WHITESPACE)) {
			int bits;

			bits = str2cidr(0,  &addr6, 0, arg);
			if (bits <= 0)
				return -1;
			memcpy(addr_bits, &addr6, sizeof(addr6));
			addr_bits[sizeof(addr6)] = bits;
			arg = "";
			passed_cidr = 1;
			passed_flags = 1;
			passed_max_clients = 1;
			passed_thold = 1;
			continue;
		}
		if (!passed_max_clients
		    && (i = strtoul(arg, &p, 10)) != 0
		    && (*p == ' ' || *p == '\t' || *p == '\0')) {
			max_clients = i;
			arg = p;
			passed_max_clients = 1;
			passed_flags = 1;
			continue;
		}
		if (!passed_thold
		    && (i = strtoul(arg, &p, 10)) > 0
		    && (*p == ' ' || *p == '\t' || *p == '\0')
		    && i <= DCC_ADMIN_CLIENTS_MAX_THOLD) {
			thold = i;
			arg = p;
			passed_thold = 1;
			passed_max_clients = 1;
			passed_flags = 1;
			continue;
		}
		help_cmd("clients", 0);
		return -1;
	}

	if (ids)
		req_flags &= ~DCC_AOP_CLIENTS_VERS;

	/* Require a server password for client IP addresses
	 * The server demands only client ID for "clients -i" */
	if (!ids
	    && !ck_cmd_priv(ce, 1, 0))
		return 0;

	if (!rdy_ctxt(0))
		return 0;

	/* Collect all of the information before printing it to minimize
	 * the changes in the position of hosts and so deleted or missing
	 * entries. */
	total = 0;
	subtotal = 0;
	max_ops = 0;
	max_nops = 0;
	memset(versions, 0, sizeof(versions));
	offset = 0;
	num_clients = 0;
	list_len = 0;
	cttop = 0;
	ctbot = 0;
	for (;;) {
		DCC_OPS result;
		int done, result_len;
		u_char vers, result_flags;
#	define BL_FLAGS (DCC_ADMN_RESP_CLIENTS_BL | DCC_ADMN_RESP_CLIENTS_BAD)
		DCC_CLNT_ID clnt_id;
		DCC_PTIME last_used;
		DCC_PTIME prev_last_used;
		DCC_SCNTR requests, nops;

		if (offset > DCC_AOP_CLIENTS_MAX_OFFSET) {
			dcc_error_msg("%d are too many clients", offset);
			break;
		}

		gettimeofday(&op_start, 0);
		result = dcc_aop(dcc_emsg, ctxt, grey_clnt_fgs,
				 NO_SRVR, clock_kludge,
				 ids ? DCC_AOP_CLIENTS_ID : DCC_AOP_CLIENTS,
				 (offset << 16) + thold,
				 ISZ(aop_resp.resp.val.string
				     ) >> DCC_ADMIN_CLIENTS_SHIFT,
				 req_flags,
				 offset >> 16,
				 addr_bits, passed_cidr ? sizeof(addr_bits) : 0,
				 &aop_resp, &op_result_su);
		if (result == DCC_OP_INVALID || result == DCC_OP_ERROR) {
			cdcc_error_msg("%s", dcc_emsg);
			break;
		}

		/* print heading before the first chunk */
		if (!offset)
			print_aop(NO_SRVR);

		result_len = (ntohs(aop_resp.hdr.len)
			      - (sizeof(aop_resp.resp)
				 - sizeof(aop_resp.resp.val.string)));
		/* stop when the server has nothing to add */
		if (result_len <= 1)
			break;

		done = 0;
		vers = 0;
		prev_last_used = 0;
		do {
			done += client_unpack(&aop_resp.resp.val.clients[done],
					      &result_flags, &clnt_id, &offset,
					      &last_used, &prev_last_used,
					      &requests, &nops, &vers, &su);
			if (result_flags & DCC_ADMN_RESP_CLIENTS_SKIP)
				continue;

			if (vers != 0) {
				if (vers < DIM(versions))
					versions[vers] += requests;
				else
					versions[0] += requests;
			}


			/* quit if we are in some kind of loop */
			if (++num_clients > DCC_ADMIN_RESP_MAX_CLIENTS)
				goto stop;
			++offset;

			/* add the new entry to the possibly sorted list */

			if (list_len <= max_clients) {
				ctnew = dcc_malloc(sizeof(*ctnew));
			} else {
				ctnew = ctbot;
				ctbot = ctbot->gt;
				if (ctbot)
					ctbot->lt = 0;
				--list_len;
			}
			memset(ctnew, 0, sizeof(*ctnew));
			ctnew->flags = (result_flags & BL_FLAGS );
			ctnew->vers = vers;
			ctnew->clnt_id = clnt_id;
			ctnew->last_used = last_used;
			ctnew->requests = requests;
			total += requests;
			if (max_ops < requests)
				max_ops = requests;
			ctnew->nops = nops;
			if (max_nops < nops)
				max_nops = nops;
			ctnew->su = su;

			ct = ctbot;
			n = list_len;
			while (ct) {
				i = !sort;
				if (!i) {
					i = (0 != (ct->flags & BL_FLAGS));
					i -= (0!=(ctnew->flags & BL_FLAGS));
				}
				if (!i)
					i = ct->requests - ctnew->requests;
				if (i >= 0)
					break;

				/* We will insert the new entry above this one
				 * in the sorted list.
				 * Increase the theshold for next batch from
				 * the server if we have enough entries */
				if (--n >= max_clients
				    && thold < ct->requests
				    && thold > 1)
					thold = min(ct->requests,
						    DCC_ADMIN_CLIENTS_MAX_THOLD);
				ct = ct->gt;
			}
			if (!ct) {
				ctnew->lt = cttop;
				if (cttop) {
					cttop->gt = ctnew;
				} else {
					ctbot = ctnew;
				}
				cttop = ctnew;
			} else {
				ctnew->gt = ct;
				ctnew->lt = ct->lt;
				if (ct->lt)
					ct->lt->gt = ctnew;
				else
					ctbot = ctnew;
				ct->lt = ctnew;
			}
			++list_len;
		} while (done < result_len);
		if (done != result_len) {
			dcc_error_msg("wrong sized clients response; %d != %d",
				      result_len, done);
			break;
		}

		/* quit if the server ran out of things to say */
		if (result_flags & DCC_ADMN_RESP_CLIENTS_LAST)
			break;

		/* Quit if we want only part of the list and we have it.
		 * We must get everything the server sends if we are sorting.
		 * The server uses our threshold to avoid sending everything
		 * it knows. */
		if (!sort && offset >= max_clients)
			break;
#undef BL_FLAGS
	}
stop:
	if (!total)
		total = 1;

	dcc_localtime(time(0), &now);

	if (max_ops > 99*1000*1000)
		ops_width = 9;
	else if (max_ops > 9*1000*1000)
		ops_width = 8;
	else
		ops_width = 7;
	if (max_nops > 99*1000)
		nops_width = 6;
	else if (max_nops > 9*1000)
		nops_width = 5;
	else
		nops_width = 4;

	/* print the list */
	num_clients = 0;
	for (ct = cttop; ct; ct = ctnew) {
		ctnew = ct->lt;
		if (num_clients == 0) {
			if (sort) {
				printf("          %*s %*s ",
				       ops_width, "ops",
				       nops_width, "nops");
				if (ids)
					fputs("   ID ", stdout);
				fputs("   last       ", stdout);
				if (req_flags & DCC_AOP_CLIENTS_VERS)
					fputs("  v", stdout);
			} else {
				printf("%*s %*s    last           ID ",
				       ops_width, "ops",
				       nops_width, "nops");
				if (req_flags & DCC_AOP_CLIENTS_VERS)
					fputs("  v", stdout);
			}
			putchar('\n');
		}
		if (++num_clients <= max_clients) {
			if (sort) {
				subtotal += ct->requests;
				printf("%3d%% %3d%% ",
				       (int)(ct->requests*100.0/total),
				       (int)(subtotal*100.0/total));
			}
			printf(L_DWPAT(*)" "L_DWPAT(*),
			       ops_width, ct->requests,
			       nops_width, ct->nops);
			if (sort && ids)
				printf(" %6d", ct->clnt_id);
			/* print year and no time if it was long ago */
			dcc_localtime(ct->last_used, &last);
			if (last.tm_year != now.tm_year
			    && (last.tm_mon < 6 || now.tm_mon > 2))
				dcc_time2str(date_buf, sizeof(date_buf),
					     " %Y/%m/%d    ", ct->last_used);
			else
				dcc_time2str(date_buf, sizeof(date_buf),
					     " %m/%d %X", ct->last_used);
			fputs(date_buf, stdout);
			if (!sort)
				printf(" %6d", ct->clnt_id);
			if (req_flags & DCC_AOP_CLIENTS_VERS) {
				if (ct->vers != 0)
					printf(" %2d", ct->vers);
				else
					fputs(" ? ", stdout);
			}
			if (ct->flags & DCC_ADMN_RESP_CLIENTS_BL)
				fputs(" BLACKLIST", stdout);
			else if (ct->flags & DCC_ADMN_RESP_CLIENTS_BAD)
				fputs(" BAD", stdout);
			if (!ids) {
				char name[DCC_MAXDOMAINLEN];
				char sustr[DCC_SU2STR_SIZE];

				if (nonames) {
					printf(" %s",
					       dcc_su2str2(sustr, sizeof(sustr),
							&ct->su));
				} else {
					printf(" %-16s %s",
					       dcc_su2str2(sustr, sizeof(sustr),
							&ct->su),
					       dcc_su2name(name, sizeof(name),
							&ct->su));
				}
			}
			putchar('\n');
		}

		memset(ct, 0, sizeof(*ct));
		dcc_free(ct);
	}
	putchar('\n');


	need_head = 1;
	for (i = 0; i < DIM(versions); ++i) {
		if (versions[i] == 0)
			continue;
		if (need_head) {
			need_head = 0;
			fputs("version    total\n", stdout);
		}
		printf("%6d "L_DWPAT(8)"\n", i, versions[i]);
	}

	return 1;
}



/* get and set the server's default anonymous client delay */
static int
anon_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	int new_delay, old_delay, inflate;
	DCC_OPS result;
	char *inflate_str, *p;

	inflate = 0;
	if (*arg == '\0') {
		new_delay = DCC_NO_ANON_DELAY;
	} else {
		if (!strcasecmp(arg, "forever")) {
			new_delay = DCC_ANON_DELAY_FOREVER;
		} else {
			new_delay = strtoul(arg, &inflate_str, 10);
			if (new_delay > DCC_ANON_DELAY_MAX
			    || (*inflate_str != '\0' && *inflate_str != ','
				&& *inflate_str != '*')) {
				dcc_error_msg("invalid delay: \"%s\"", arg);
				return 0;
			}
			if (*inflate_str != '\0') {
				++inflate_str;
				inflate_str += strspn(inflate_str,
						      DCC_WHITESPACE);
			}
			if (*inflate_str != '\0'
			    && strcasecmp(inflate_str, "none")) {
				inflate = strtoul(inflate_str, &p, 10);
				if (*p != '\0') {
					dcc_error_msg("invalid delay inflation:"
						      " \"%s\"", inflate_str);
					return 0;
				}
			}
		}
		if (!ck_cmd_priv(ce, 1, 0))
			return 0;
	}

	if (!rdy_ctxt(0))
		return 0;

	gettimeofday(&op_start, 0);
	result = dcc_aop(dcc_emsg, ctxt, grey_clnt_fgs,
			 NO_SRVR, clock_kludge, DCC_AOP_ANON_DELAY,
			 inflate, new_delay>>8, new_delay, 0, 0, 0,
			 &aop_resp, &op_result_su);
	if (result == DCC_OP_INVALID
	    || result == DCC_OP_ERROR) {
		cdcc_error_msg("%s", dcc_emsg);
		return 0;
	}

	old_delay = ((aop_resp.resp.val.anon_delay.delay[0]<<8)
		     + aop_resp.resp.val.anon_delay.delay[1]);
	if (old_delay == DCC_ANON_DELAY_FOREVER) {
		printf("    anon delay %s FOREVER\n",
		       new_delay != DCC_NO_ANON_DELAY ? "was" : "is");
	} else {
		printf("    anon delay %s %d",
		       new_delay != DCC_NO_ANON_DELAY ? "was" : "is",
		       old_delay);
		inflate = ((aop_resp.resp.val.anon_delay.inflate[0]<<24)
			   +(aop_resp.resp.val.anon_delay.inflate[1]<<16)
			   +(aop_resp.resp.val.anon_delay.inflate[2]<<8)
			   +aop_resp.resp.val.anon_delay.inflate[3]);
		if (inflate != 0)
			printf(",%d", inflate);
		putchar('\n');
	}
	return 1;
}



/* rewind the flood from a single server */
static int
flod_rewind(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_CLNT_ID id;

	if (!arg)
		return -1;
	id = dcc_get_id(0, arg, 0, 0);
	if (id == DCC_ID_INVALID)
		return -1;

	return do_aop(DCC_AOP_FLOD, id*256 + DCC_AOP_FLOD_REWIND, NO_SRVR, 1);
}



/* fast forward the flood to a single server */
static int
ffwd_out(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_CLNT_ID id;

	if (!arg)
		return -1;
	id = dcc_get_id(0, arg, 0, 0);
	if (id == DCC_ID_INVALID)
		return -1;

	return do_aop(DCC_AOP_FLOD, id*256 + DCC_AOP_FLOD_FFWD_OUT, NO_SRVR, 1);
}



/* fast forward the flood to a single server */
static int
ffwd_in(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_CLNT_ID id;

	if (!arg)
		return -1;
	id = dcc_get_id(0, arg, 0, 0);
	if (id == DCC_ID_INVALID)
		return -1;

	return do_aop(DCC_AOP_FLOD, id*256 + DCC_AOP_FLOD_FFWD_IN, NO_SRVR, 1);
}



/* get the flood counts for a server */
static int
do_flod_stats(const char *arg, DCC_AOP_FLODS op)
{
	u_int32_t id, next_id;
	u_char heading;
	int sresult;

	if (!arg)
		return -1;

	heading = 1;
	if (!strcasecmp(arg, "all")) {
		id = DCC_SRVR_ID_MAX+1;
		for (;;) {
			if (!start_aop(DCC_AOP_FLOD, id*256 + op, NO_SRVR))
				return 0;
			sresult = sscanf(aop_resp.resp.val.string,
					 DCC_AOP_FLOD_STATS_ID, &next_id);
			if (1 == sresult
			    && id == next_id) {
				if (id == DCC_SRVR_ID_MAX+1) {
					BUFCPY(aop_resp.resp.val.string,
					       " (no flooding peers)");
					fin_aop(NO_SRVR, 1);
				}
				return 1;
			}
			fin_aop(NO_SRVR, heading);
			heading = 0;
			if (1 != sresult)
				return 0;
			id = next_id+DCC_SRVR_ID_MAX+1;
		}
	}

	id = dcc_get_id(0, arg, 0, 0);
	if (id == DCC_ID_INVALID)
		return -1;
	return do_aop(DCC_AOP_FLOD, id*256 + op, NO_SRVR, heading);
}



static int
flod_stats(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	return do_flod_stats(arg, DCC_AOP_FLOD_STATS);
}



static int
flod_stats_clear(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	return do_flod_stats(arg, DCC_AOP_FLOD_STATS_CLEAR);
}



/* get the statistics from all known servers */
static int
stats_cmd(const char *arg, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	DCC_SRVR_CLASS *class;
	SRVR_INX srvr_inx;
	int srvrs_gen;
	DCC_AOPS aop;

	/* look for "clear" or "clear all" */
	srvr_inx = NO_SRVR;
	aop = DCC_AOP_STATS;
	while (*arg != 0) {
		arg += strspn(arg, " \t");
		if (srvr_inx == NO_SRVR
		    && !CLITCMP(arg, "clear")) {
			arg += LITZ("clear");
			aop = DCC_AOP_STATS_CLEAR;
			if (!get_passwd(aops_tbl[aop].privileged)) {
				dcc_error_msg("\"stats clear\""
					      " is a privileged operation"
					      PRV_MSG, ids_path);
				return 0;
			}
		} else if (aop == DCC_AOP_STATS
			   && !CLITCMP(arg, "all")) {
			arg += LITZ("all");
			srvr_inx = 0;
		}
		if (*arg != '\0' && *arg != ' ' && *arg != '\t')
			return -1;
	}

	if (!rdy_ctxt(0))
		return 0;
	class = DCC_GREY2CLASS(grey_on);
	srvrs_gen = class->gen;
	do {
		if (srvrs_gen != class->gen) {
			dcc_error_msg("list of servers changed");
			return 0;
		}
		/* skip servers that have never answered a NOP */
		if (srvr_inx != NO_SRVR
		    && class->addrs[srvr_inx].srvr_id == DCC_ID_INVALID )
			continue;

		do_aop(aop, sizeof(aop_resp.resp.val.string), srvr_inx, 1);
		fflush(stderr);
		fflush(stdout);
	} while (srvr_inx != NO_SRVR && ++srvr_inx < class->num_srvrs);

	return 1;
}



static int
clock_ck_cmd(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (!rdy_ctxt(0))
		return 0;
	do_aop(DCC_AOP_CLOCK_CHECK, sizeof(aop_resp.resp.val.string),
	       NO_SRVR, 1);

	return 1;
}



/* restore tracing to default */
static int
trace_def(const char *arg DCC_UNUSED, const CMD_TBL_ENTRY *ce DCC_UNUSED)
{
	if (!rdy_ctxt(0))
		return 0;

	return (do_aop(DCC_AOP_TRACE_ON,
		       grey_on ? DCC_TRACE_GREY_DEF : DCC_TRACE_DEF,
		       NO_SRVR, 1)
		&& do_aop(DCC_AOP_TRACE_OFF,
			  DCC_TRACE_BITS
			  & ~(grey_on ? DCC_TRACE_GREY_DEF : DCC_TRACE_DEF),
			  NO_SRVR, 1));
}
